# -*- coding: utf-8 -*-

from mq_http_sdk.mq_client import MQClient as _MQClient
from mq_http_sdk.mq_consumer import MQConsumer as _MQConsumer
from mq_http_sdk.mq_producer import MQProducer as _MQProducer, MQTransProducer as _MQTransProducer

from hagworm.extend.asyncio.future import ThreadPool


class MQConsumer(_MQConsumer):

    async def consume_message(self, batch_size=1, wait_seconds=-1):

        return await self.mq_client.thread_pool.run(super().consume_message, batch_size, wait_seconds)

    async def ack_message(self, receipt_handle_list):

        return await self.mq_client.thread_pool.run(super().ack_message, receipt_handle_list)


class MQProducer(_MQProducer):

    async def publish_message(self, message):

        return await self.mq_client.thread_pool.run(super().publish_message, message)


class MQTransProducer(_MQTransProducer):

    async def consume_half_message(self, batch_size=1, wait_seconds=-1):

        return await self.mq_client.thread_pool.run(super().consume_half_message, batch_size, wait_seconds)

    async def commit(self, receipt_handle):

        return await self.mq_client.thread_pool.run(super().commit, receipt_handle)

    async def rollback(self, receipt_handle):

        return await self.mq_client.thread_pool.run(super().rollback, receipt_handle)


class MQClient(_MQClient):

    def __init__(self, host, access_id, access_key, security_token=r'', debug=False, logger=None):

        super().__init__(host, access_id, access_key, security_token, debug, logger)

        # 原版SDK中的MQClient线程不安全，并利用线程池仅有一个线程这种特例，保证线程安全
        self._thread_pool = ThreadPool(1)

    @property
    def thread_pool(self):

        return self._thread_pool

    def get_consumer(self, instance_id, topic_name, consumer, message_tag=r''):

        return MQConsumer(instance_id, topic_name, consumer, message_tag, self, self.debug)

    def get_producer(self, instance_id, topic_name):

        return MQProducer(instance_id, topic_name, self, self.debug)

    def get_trans_producer(self, instance_id, topic_name, group_id):

        return MQTransProducer(instance_id, topic_name, group_id, self, self.debug)



# -*- coding: utf-8 -*-

from cacheout import LRUCache

from hagworm.extend.base import Utils
from hagworm.extend.transaction import Transaction


class StackCache:
    """堆栈缓存

    使用运行内存作为高速缓存，可有效提高并发的处理能力

    """

    def __init__(self, maxsize=0xffff, ttl=None):

        self._cache = LRUCache(maxsize, ttl)

    def has(self, key):

        return self._cache.has(key)

    def get(self, key, default=None):

        return self._cache.get(key, default)

    def set(self, key, val, ttl=None):

        self._cache.set(key, val, ttl)

    def incr(self, key, val=1, ttl=None):

        res = self._cache.get(key, 0) + val

        self._cache.set(key, res, ttl)

        return res

    def decr(self, key, val=1, ttl=None):

        res = self._cache.get(key, 0) - val

        self._cache.set(key, res, ttl)

        return res

    def delete(self, key):

        return self._cache.delete(key)

    def size(self):

        return self._cache.size()


class PeriodCounter:

    MIN_EXPIRE = 60

    def __init__(self, time_slice, key_prefix=r'', maxsize=0xffff):

        self._time_slice = time_slice
        self._key_prefix = key_prefix

        # 缓存对象初始化，key最小过期时间60秒
        self._cache = StackCache(maxsize, max(time_slice, self.MIN_EXPIRE))

    def _get_key(self, key=None):

        time_period = Utils.math.floor(Utils.timestamp() / self._time_slice)

        if key is None:
            return f'{self._key_prefix}_{time_period}'
        else:
            return f'{self._key_prefix}_{key}_{time_period}'

    def get(self, key=None):

        _key = self._get_key(key)

        return self._cache.get(_key, 0)

    def incr(self, val, key=None):

        _key = self._get_key(key)

        return self._cache.incr(_key, val)

    def incr_with_trx(self, val, key=None):

        _key = self._get_key(key)

        trx = Transaction()
        trx.add_rollback_callback(self._cache.decr, _key, val)

        return self._cache.incr(_key, val), trx

    def decr(self, val, key=None):

        _key = self._get_key(key)

        return self._cache.decr(_key, val)

    def decr_with_trx(self, val, key=None):

        _key = self._get_key(key)

        trx = Transaction()
        trx.add_rollback_callback(self._cache.incr, _key, val)

        return self._cache.decr(_key, val), trx



# -*- coding: utf-8 -*-

import zmq

from concurrent.futures import CancelledError
from zmq.asyncio import Context

from hagworm.extend.base import ContextManager
from hagworm.extend.asyncio.base import Utils, AsyncCirculator
from hagworm.extend.asyncio.buffer import QueueBuffer


class _SocketBase(ContextManager):

    def __init__(self, name, socket_type, address, bind_mode):

        self._name = name if name else Utils.uuid1()[:8]

        self._context = Context.instance()
        self._socket = self._context.socket(socket_type)

        self._address = address
        self._bind_mode = bind_mode

    def _context_initialize(self):

        self.open()

    def _context_release(self):

        self.close()

    def open(self):

        if self._bind_mode:
            self._socket.bind(self._address)
        else:
            self._socket.connect(self._address)

    def close(self):

        if not self._socket.closed:
            self._socket.close()

    def set_hwm(self, val):

        self._socket.set_hwm(val)


class Subscriber(_SocketBase):

    def __init__(self, address, bind_mode=False, *, name=None, topic=r''):

        super().__init__(name, zmq.SUB, address, bind_mode)

        self._msg_listen_task = None

        self._socket.setsockopt_string(zmq.SUBSCRIBE, topic)

    async def _message_listener(self):

        while not self._socket.closed:
            try:
                await self._message_handler(
                    await self._socket.recv_pyobj()
                )
            except CancelledError as _:
                pass
            except Exception as err:
                Utils.log.error(err)

    async def _message_handler(self, data):

        raise NotImplementedError()

    def open(self):

        super().open()

        self._msg_listen_task = Utils.create_task(self._message_listener())

    def close(self):

        super().close()

        if self._msg_listen_task is not None:
            self._msg_listen_task.cancel()


class Publisher(_SocketBase):

    def __init__(self, address, bind_mode=False, *, name=None):

        super().__init__(name, zmq.PUB, address, bind_mode)

    async def send(self, data):

        await self._socket.send_pyobj(data)


class PublisherWithBuffer(_SocketBase, QueueBuffer):

    def __init__(self, address, bind_mode=False, *, name=None, buffer_maxsize=0xffff, buffer_timeout=1):

        _SocketBase.__init__(self, name, zmq.PUB, address, bind_mode)
        QueueBuffer.__init__(self, buffer_maxsize, buffer_timeout)

    async def _run(self, data_list):

        await self._socket.send_pyobj(data_list)

    async def safe_close(self, timeout=0):

        async for _ in AsyncCirculator(timeout):
            if len(self._data_list) == 0:
                super().close()
                break



# -*- coding: utf-8 -*-

import os
import sys

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../'))

from hagworm.frame.tornado.base import Launcher

from routing import router
from setting import ConfigStatic, ConfigDynamic
from service.base import GlobalDict, DataSource


def main():

    cluster = os.getenv(r'CLUSTER', None)

    if cluster is None:
        ConfigStatic.read(r'./static.conf')
        ConfigDynamic.read(r'./dynamic.conf')
    else:
        ConfigStatic.read(f'./static.{cluster.lower()}.conf')
        ConfigDynamic.read(f'./dynamic.{cluster.lower()}.conf')

    # 初始化全局变量字典（进程间同步）
    GlobalDict()

    Launcher(
        router,
        ConfigDynamic.Port,
        process_num=ConfigDynamic.ProcessNum,
        async_initialize=DataSource.initialize,
        debug=ConfigDynamic.Debug,
        gzip=ConfigDynamic.GZip,
        template_path=r'view',
        cookie_secret=ConfigDynamic.Secret,
        log_level=ConfigDynamic.LogLevel,
        log_file_path=ConfigDynamic.LogFilePath,
        log_file_num_backups=ConfigDynamic.LogFileBackups,
    ).start()


if __name__ == r'__main__':

    main()
